package fr.takima.lotka_volterra;

import fr.takima.lotka_volterra.common.value.PositiveInt;
import fr.takima.lotka_volterra.moving.MapConfigurer;
import fr.takima.lotka_volterra.moving.Position;
import fr.takima.lotka_volterra.moving.PositionComponent;
import fr.takima.lotka_volterra.predator.Predator;
import fr.takima.lotka_volterra.predator.PredatorIdGenerator;
import fr.takima.lotka_volterra.predator.Satiety;
import fr.takima.lotka_volterra.prey.Prey;
import fr.takima.lotka_volterra.prey.PreyIdGenerator;
import fr.takima.lotka_volterra.simulation.Simulation;
import fr.takima.lotka_volterra.simulation.SimulationConfigurer;
import fr.takima.lotka_volterra.simulation.step.Step;
import fr.takima.lotka_volterra.simulation.step.Steps;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.CommandLineRunner;
import org.springframework.boot.SpringApplication;
import org.springframework.boot.autoconfigure.SpringBootApplication;

import java.util.HashSet;
import java.util.Set;

@SpringBootApplication
public class EntryPoint implements CommandLineRunner {

    private static final Logger log = LoggerFactory.getLogger(EntryPoint.class);

    public static void main(String[] args) {
        SpringApplication.run(EntryPoint.class, args);
    }

    private final SimulationConfigurer simulationConfigurer;
    private final PreyIdGenerator preyIdGenerator;
    private final PredatorIdGenerator predatorIdGenerator;

    public EntryPoint(SimulationConfigurer simulationConfigurer,
                      PreyIdGenerator preyIdGenerator,
                      PredatorIdGenerator predatorIdGenerator) {
        this.simulationConfigurer = simulationConfigurer;
        this.preyIdGenerator = preyIdGenerator;
        this.predatorIdGenerator = predatorIdGenerator;
    }

    @Override
    public void run(String... args) throws Exception {
        var preys = getInitialPreys(simulationConfigurer.initialPreyCount(), simulationConfigurer.map());
        var predators = getInitialPredators(simulationConfigurer.initialPredatorCount(), simulationConfigurer.map());
        var step = new Step(new Step.Number(0), preys, predators);
        var simulation = new Simulation(simulationConfigurer, new Steps(Set.of(step)));
        run(simulation, simulationConfigurer.stepCount());
    }

    private void run(Simulation simulation, PositiveInt stepCount) {
        for (var i = 0; i < stepCount.value(); i++) {
            simulation = simulation.runNextStep();
        }
        var populations = simulation.steps().last().populations();
        log.info("preys: {}, predators: {}", populations.preys().value(), populations.predators().value());
    }

    private Set<Prey> getInitialPreys(PositiveInt preyCount, MapConfigurer configurer) {
        var preys = new HashSet<Prey>();

        for (var i = 0; i < preyCount.value(); i++) {
            var preyId = preyIdGenerator.generate();
            var x = new PositionComponent(0, configurer.minX(), configurer.maxX());
            var y = new PositionComponent(0, configurer.minY(), configurer.maxY());
            var position = new Position(x, y);
            var prey = new Prey(preyId, position);
            preys.add(prey);
        }

        return preys;
    }

    private Set<Predator> getInitialPredators(PositiveInt predatorCount, MapConfigurer configurer) {
        var predators = new HashSet<Predator>();

        for (var i = 0; i < predatorCount.value(); i++) {
            var predatorId = predatorIdGenerator.generate();
            var x = new PositionComponent(0, configurer.minX(), configurer.maxX());
            var y = new PositionComponent(0, configurer.minY(), configurer.maxY());
            var position = new Position(x, y);
            var predator = new Predator(predatorId, Satiety.MAX, position);
            predators.add(predator);
        }

        return predators;
    }

}
