package fr.takima.lotka_volterra.mating;

import fr.takima.lotka_volterra.validator.NotNullValidator;

import java.math.BigDecimal;
import java.util.HashSet;
import java.util.Set;
import java.util.stream.Collectors;

import static java.util.Collections.unmodifiableSet;

public final class Mating<T extends MatingEntity<T>> {

    private static final int MAX_SUCCESS_THRESHOLD_INT = 100;
    private static final BigDecimal MAX_SUCCESS_THRESHOLD_BIG_DECIMAL = new BigDecimal(MAX_SUCCESS_THRESHOLD_INT);

    private final MatingConfigurer configurer;
    private final Set<T> matingEntities;
    private final ChildMaking<T> childMaking;
    private final BigDecimal successThreshold;

    public Mating(MatingConfigurer configurer, Set<T> matingEntities, ChildMaking<T> childMaking) {
        new NotNullValidator("configurer", configurer).validate();
        new NotNullValidator("matingEntities", matingEntities).validate();
        new NotNullValidator("childMaking", childMaking).validate();
        this.configurer = configurer;
        this.matingEntities = matingEntities;
        this.childMaking = childMaking;
        this.successThreshold = getSuccessThreshold();
    }

    public Set<T> run() {
        var couples = new HashSet<Couple<T>>();
        var matingCandidates = new HashSet<>(matingEntities);
        for (var entity : matingEntities) {
            var isCandidate = matingCandidates.contains(entity);
            var iterator = matingCandidates.iterator();
            var nextMatingCandidates = new HashSet<T>();
            while (isCandidate && iterator.hasNext()) {
                var other = iterator.next();
                if (!entity.equals(other)) {
                    if (entity.isWithin(configurer.maxDistance(), other)) {
                        var matingSucceeded = computeSuccess();
                        if (matingSucceeded) {
                            couples.add(new Couple<>(entity, other));
                            isCandidate = false;
                        } else {
                            nextMatingCandidates.add(other);
                        }
                    } else {
                        nextMatingCandidates.add(other);
                    }
                }
            }
            matingCandidates = nextMatingCandidates;
        }

        var children = couples.stream()
                .map(childMaking::run)
                .flatMap(Set::stream)
                .collect(Collectors.toSet());

        var next = new HashSet<T>();
        next.addAll(matingEntities);
        next.addAll(children);

        return unmodifiableSet(next);
    }

    private boolean computeSuccess() {
        var rand = new BigDecimal(configurer.random().nextInt(MAX_SUCCESS_THRESHOLD_INT));
        return rand.compareTo(successThreshold) < 0;
    }

    private BigDecimal getSuccessThreshold() {
        var successProbabilityValue = configurer.successProbability().value();
        return successProbabilityValue.multiply(MAX_SUCCESS_THRESHOLD_BIG_DECIMAL);
    }

}
