package fr.takima.lotka_volterra.predation;

import fr.takima.lotka_volterra.predator.Predator;
import fr.takima.lotka_volterra.prey.Prey;
import fr.takima.lotka_volterra.validator.NotNullValidator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.math.BigDecimal;
import java.util.HashSet;
import java.util.Set;

public final class Predation {

    private static final Logger log = LoggerFactory.getLogger(Predation.class);
    private static final int MAX_SUCCESS_THRESHOLD_INT = 100;
    private static final BigDecimal MAX_SUCCESS_THRESHOLD_BIG_DECIMAL = new BigDecimal(MAX_SUCCESS_THRESHOLD_INT);

    private final PredationConfigurer configurer;
    private final Set<Prey> preys;
    private final Set<Predator> predators;
    private final BigDecimal successThreshold;

    public Predation(PredationConfigurer configurer, Set<Prey> preys, Set<Predator> predators) {
        new NotNullValidator("configurer", configurer).validate();
        new NotNullValidator("preys", preys).validate();
        new NotNullValidator("predators", predators).validate();
        this.configurer = configurer;
        this.preys = preys;
        this.predators = predators;
        this.successThreshold = getSuccessThreshold();
    }

    public PredationResult run() {
        var nextPredators = new HashSet<Predator>();
        var nextPreys = new HashSet<>(preys);

        for (var predator : predators) {
            var remainingPreys = new HashSet<Prey>();
            var alreadyAte = false;
            for (var prey : nextPreys) {
                if (!alreadyAte && predator.isWithin(configurer.maxDistance(), prey)) {
                    var predationSucceeded = computeSuccess();
                    if (predationSucceeded) {
                        var updatedPredator = predator.withSatietyGain(configurer.successSatietyGain());
                        nextPredators.add(updatedPredator);
                        alreadyAte = true;
                        log.info("predator '{}' ate prey '{}'", predator.id().value(), prey.id().value());
                    } else {
                        remainingPreys.add(prey);
                    }
                } else {
                    remainingPreys.add(prey);
                }
            }

            if (!alreadyAte) {
                var updatedPredator = predator.withSatietyLoss(configurer.failureSatietyLoss());
                if (!updatedPredator.starvedToDeath()) {
                    nextPredators.add(updatedPredator);
                } else {
                    log.info("predator '{}' starved to death", predator.id().value());
                }
            }

            nextPreys = remainingPreys;
        }

        return new PredationResult(nextPreys, nextPredators);
    }

    private boolean computeSuccess() {
        var rand = new BigDecimal(configurer.random().nextInt(MAX_SUCCESS_THRESHOLD_INT));
        return rand.compareTo(successThreshold) < 0;
    }

    private BigDecimal getSuccessThreshold() {
        var successProbabilityValue = configurer.successProbability().value();
        return successProbabilityValue.multiply(MAX_SUCCESS_THRESHOLD_BIG_DECIMAL);
    }

}
