package fr.takima.lotka_volterra.predator;

import fr.takima.lotka_volterra.moving.Position;
import fr.takima.lotka_volterra.moving.PositionComponent;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;

import static org.assertj.core.api.Assertions.assertThat;

class PredatorTest {

    private static final Position POSITION_0_0 = new Position(new PositionComponent(0, -100, 100), new PositionComponent(0, -100, 100));
    private static final Position POSITION_1_1 = new Position(new PositionComponent(1, -100, 100), new PositionComponent(1, -100, 100));
    private static final Predator ENTITY_1 = new Predator(new PredatorId("1"), new Satiety(50), POSITION_0_0);
    private static final Predator ENTITY_1_WITH_SATIETY_GAIN = new Predator(new PredatorId("1"), Satiety.MAX, POSITION_0_0);
    private static final Predator ENTITY_1_WITH_SATIETY_LOSS = new Predator(new PredatorId("1"), Satiety.MIN, POSITION_0_0);
    private static final Predator ENTITY_1_WITH_POSITION = new Predator(new PredatorId("1"), Satiety.MAX, POSITION_1_1);
    private static final Predator ENTITY_2 = new Predator(new PredatorId("2"), Satiety.MAX, POSITION_0_0);

    @Nested
    class WithSatietyGainTest {

        @Test
        void nominal() {
            var actual = ENTITY_1.withSatietyGain(new Satiety(50));
            assertThat(actual).usingRecursiveComparison().isEqualTo(ENTITY_1_WITH_SATIETY_GAIN);
        }

        @Test
        void above_max() {
            var actual = ENTITY_1.withSatietyGain(new Satiety(51));
            assertThat(actual).usingRecursiveComparison().isEqualTo(ENTITY_1_WITH_SATIETY_GAIN);
        }

    }

    @Nested
    class WithSatietyLossTest {

        @Test
        void nominal() {
            var actual = ENTITY_1.withSatietyLoss(new Satiety(50));
            assertThat(actual).usingRecursiveComparison().isEqualTo(ENTITY_1_WITH_SATIETY_LOSS);
        }

        @Test
        void under_min() {
            var actual = ENTITY_1.withSatietyLoss(new Satiety(51));
            assertThat(actual).usingRecursiveComparison().isEqualTo(ENTITY_1_WITH_SATIETY_LOSS);
        }

    }

    @Nested
    class StarvedToDeathTest {

        @Test
        void _false() {
            var actual = ENTITY_1.starvedToDeath();
            assertThat(actual).isFalse();
        }

        @Test
        void _true() {
            var actual = ENTITY_1_WITH_SATIETY_LOSS.starvedToDeath();
            assertThat(actual).isTrue();
        }

    }

    @Nested
    class WithPositionTest {

        @Test
        void ok() {
            var actual = ENTITY_1.withPosition(POSITION_1_1);
            assertThat(actual).isEqualTo(ENTITY_1_WITH_POSITION);
        }

    }

    @Nested
    class EqualsTest {

        @Test
        void with_itself_returns_true() {
            var actual = ENTITY_1.equals(ENTITY_1);
            assertThat(actual).isTrue();
        }

        @Test
        void with_null_returns_false() {
            var actual = ENTITY_1.equals(null);
            assertThat(actual).isFalse();
        }

        @Test
        void with_other_class_returns_null() {
            var actual = ENTITY_1.equals("");
            assertThat(actual).isFalse();
        }

        @Test
        void with_other_id_returns_false() {
            var actual = ENTITY_1.equals(ENTITY_2);
            assertThat(actual).isFalse();
        }

        @Test
        void with_same_id_returns_true() {
            var actual = ENTITY_1.equals(ENTITY_1_WITH_POSITION);
            assertThat(actual).isTrue();
        }

    }

}
